// Includes
#include <Python.h>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#ifdef _WIN32
#  include <windows.h>
#  include <winsock.h>
#endif
#include <list>
#include <vector>
#include <map>
#ifdef __unix__
#  include <signal.h>
#  include <unistd.h>
#endif
#include <time.h>

#include "NetSock.h"

// Namespace
using namespace std;

// Defines
#define SILK_VERSION "0.0.1"
#define UNUSED (void)
#define ERROR_PTR (void*)(-1) // Stupid hack
#define SILK_UDP 1
#define SILK_TCP 0
#define SILK_PASS 0
#define SILK_INT  1
#define SILK_LOG  2

// This are not config settings ;D
#define SILK_SERVER false
#define SILK_CLIENT true

// Structures
struct PacketLog
{
  unsigned int Type;       // 0 == S->C; 1 == C->S
  unsigned int OrgLen;     // Original Length

  bool OutPacketIncluded;  // True or false
  unsigned int OutLen;     // Out Length

  // Followed by:
  // - BYTE OriginalPacket[OrgLen]
  // - BYTE OutPacket[OutLen] if OutPacketIncluded
};

struct SockPair
{
  unsigned int    LastDataFromClient; // TODO Use this
  unsigned int    LastDataFromServer; // TODO Use this
  NetSock *Client;
  NetSock *Server;
  FILE    *Log;
  int      Handle;

  int      ServHandle;  
  int      Type;

  // UDP Only stuff
  char     UDPDstHost[256];
  int      UDPDstPort;

  char     UDPLastSrcHost[256];
  int      UDPLastSrcPort;
};

// TODO Add UDP support here
struct SockServ
{
  int         Handle;
  NetSock    *Server;
  const char *DstHost;
  int         DstPort;
  int         Type;
};

// Replay files/list descriptor
// A handle to this is returned by PyAPI silk.ReplayLoad
// and used by the other silk.Replay* functions
// TODO Add option to open replays of the currently tunneled transmitions
struct ReplayDesc
{
  // Handle to the replay file
  // This handle is NULL if cache is active
  FILE *FileHandle;

  // Cache flag, if true, then the whole
  // file content is loaded to the memory
  // This is a lot faster then reading a packet
  // on demand, but in case of large logs,
  // you have lots of free memory
  bool CacheActive;

  // Vector of packet logs
  vector<PacketLog> CachedFile;
};

// Globals
namespace Globals
{
  // Interactive Mode?
  bool InteractiveMode;

  // The end?
  bool TheEnd;

  // Handle counter
  int HandleCounter;
  int ConnHandleCounter;

  // SockPair list
  list<SockPair*> TCPConnections;
  list<SockPair*> UDPConnections;
  list<SockServ*> TCPServers;

  // Open replay files map
  map<int, ReplayDesc*> ReplayMaps;
  int ReplayHandleCounter;

  // Script name and handle
  // Both of this are null if there is no script
  // and it should run in interactive mode
  char *ScriptName;
  PyObject *PyScript;

  void *CurrentPacketData;
  int   CurrentPacketLen;
};

// Functions
#ifdef _WIN32
bool InitWinsock();
#endif
bool InitPython();
char* xstrdup(const char *ExistingString);
PyObject* InvokeFunc(PyObject *PyScript, const char *FuncName, PyObject *PyArgs = NULL);
PyObject* ImportScript(const char *ScriptName);

void InteractiveMode();
void ProxyMode();
void PrintHex(void *pdata, int len, int from, int to);

#ifdef __unix__
void HandleCtrlC(int SigNo);
#else
BOOL WINAPI HandleCtrlC(DWORD CtrlType);
#endif

bool HandleAcceptTCP(NetSock *NewClient, SockServ *Serv);
bool HandleTunnel(SockPair *Pair, int Family, bool Client);

int CreateTCPProxy(int Type,
  char *ServBindHost, int ServBindPort,
  char *DstHost, int DstPort);

int CreateUDPProxy(int Type,
  char* SrcBindHost, int SrcBindPort,
  char* DefaultSrcHost, int DefaultSrcPort,
  char* DstHost, int DstPort, 
  char* DstBindHost, int DstBindPort);

// API
PyObject* SilkAPI_Version(PyObject *self, PyObject *args);
PyObject* SilkAPI_UDP(PyObject *self, PyObject *args);
PyObject* SilkAPI_TCP(PyObject *self, PyObject *args);
PyObject* SilkAPI_PASS(PyObject *self, PyObject *args);
PyObject* SilkAPI_INT(PyObject *self, PyObject *args);
PyObject* SilkAPI_LOG(PyObject *self, PyObject *args);
PyObject* SilkAPI_Go(PyObject *self, PyObject *args);
PyObject* SilkAPI_Stop(PyObject *self, PyObject *args);
PyObject* SilkAPI_Create(PyObject *self, PyObject *args);
PyObject* SilkAPI_Destroy(PyObject *self, PyObject *args);
PyObject* SilkAPI_Exit(PyObject *self, PyObject *args);
PyObject* SilkAPI_PrintHex(PyObject *self, PyObject *args);
PyObject* SilkAPI_Fuzz(PyObject *self, PyObject *args);
PyObject* SilkAPI_Rand(PyObject *self, PyObject *args);
PyObject* SilkAPI_ReplayLoad(PyObject *self, PyObject *args);
PyObject* SilkAPI_ReplayClose(PyObject *self, PyObject *args);
PyObject* SilkAPI_ReplayGetPacket(PyObject *self, PyObject *args);

static PyMethodDef SilkPythonAPI[] = 
{
  // Initial
  { "Version", SilkAPI_Version, METH_NOARGS, "Returns the SilkProxy version." },
  { "Exit", SilkAPI_Exit, METH_NOARGS, "Quits." },  
  { "Quit", SilkAPI_Exit, METH_NOARGS, "Quits." },    
  { "Bye",  SilkAPI_Exit, METH_NOARGS, "Quits." },

  { "UDP", SilkAPI_UDP, METH_NOARGS, "Returns int(1)." },
  { "TCP", SilkAPI_TCP, METH_NOARGS, "Returns int(0)." },

  { "PASS", SilkAPI_PASS, METH_NOARGS, "Returns int(0)." },
  { "INTERRUPT", SilkAPI_INT, METH_NOARGS, "Returns int(1)." },  
  { "INT", SilkAPI_INT, METH_NOARGS, "Returns int(1)." },    
  { "LOG", SilkAPI_LOG, METH_NOARGS, "Returns int(2)." },

  { "Fuzz", SilkAPI_Fuzz, METH_VARARGS, "Simple fuzzer." },
  { "Rand", SilkAPI_Rand, METH_VARARGS, "Random(range)." },  

  { "PrintHex", SilkAPI_PrintHex, METH_VARARGS, "Prints packet in human-readable form." },

  { "Go", SilkAPI_Go, METH_NOARGS, "After python event-handler finishes, it will return to proxy-mode until breakpoint." },
  { "Stop", SilkAPI_Stop, METH_NOARGS, "After python event-handler finishes, it will return to interactive-mode." },

  { "Create", SilkAPI_Create, METH_VARARGS, "Create a tunnel/proxy. Usage: Create(UDP() or TCP(), (bindip srcport), (destip srcport), PASS() or INTERRUPT() or LOG())" },
  { "Destroy", SilkAPI_Destroy, METH_VARARGS, "Destroy a tunnel/proxy. Usage: Destroy(handle)" },  

  { "ReplayLoad", SilkAPI_ReplayLoad, METH_VARARGS, "Loads a replay from a file" },
  { "ReplayOpen", SilkAPI_ReplayLoad, METH_VARARGS, "Loads a replay from a file" },
  { "ReplayClose", SilkAPI_ReplayClose, METH_VARARGS, "Close the replay" },
  { "ReplayGetPacket", SilkAPI_ReplayGetPacket, METH_VARARGS, "Get packet from replay" },
  // TODO Add time-related packet queries
  
  // Done
  { NULL, NULL, 0, NULL }
};


// Main
int
main(int argc, char **argv)
{
  // Display banner
  puts("SilkProxy v." SILK_VERSION " by gynvael.coldwind//vx");
  srand(time(NULL));

  // Winsock if needed
#ifdef _WIN32
  if(!InitWinsock())
    return 1;
#endif

  // Check arguments
  if(argc != 1)
  {
    // The second argument is a script name
    Globals::ScriptName = xstrdup(argv[1]);

    // Check the existance of '.py'
    char *p = strstr(Globals::ScriptName, ".py");
    if(p && p[3] == '\0')
    {
      // Remove the .py, since the python lib doesn't like it
      *p = '\0';
    }
  }

  // Init signal stuff
#ifdef __unix__
  signal(SIGINT, HandleCtrlC);
#else
  SetConsoleCtrlHandler(HandleCtrlC, TRUE);
#endif

  // Set interactive mode
  // Please note that the script may change this by running silk.Go()
  Globals::InteractiveMode = true;

  // Init the python script stuff
  if(!InitPython())
    return 2;

  // Import the user-script and run Silk_OnInit
  if(Globals::ScriptName)
  {
    // Import the script
    Globals::PyScript = ImportScript(Globals::ScriptName);
    if(!Globals::PyScript)
      return 3;

    // Invoke the Silk_OnInit
    PyObject *PyRet = InvokeFunc(Globals::PyScript, "Silk_OnInit");
    if((void*)PyRet == ERROR_PTR)
    {
      Py_DECREF(Globals::PyScript);
      return 4;
    }
    Py_XDECREF(PyRet);
  }

  // Enter the main loop
  while(!Globals::TheEnd)
  {
    // Interactive mode? Then wait for commands
    if(Globals::InteractiveMode)
    {
      // Interactive mode until someone says otherwise
      // While this is active, the tunnel does not acctually work
      InteractiveMode();
    }
    else
    {
      // Run mode
      ProxyMode();
    }
  }

  // Get rid of script
  Py_XDECREF(Globals::PyScript);  

  // Done
  return 0;
}


#ifdef _WIN32
bool
InitWinsock()
{
  WSADATA wsdat;
  memset(&wsdat,0,sizeof(wsdat));

  if(WSAStartup(0x0101,&wsdat))
  {
    printf("WSAStartup()\n");
    return false;
  }

  return true;
}
#endif

char*
xstrdup(const char *ExistingString)
{
  char *p = strdup(ExistingString);
  if(!p)
  {
    fprintf(stderr, "Error: not enough memory (xstrdup)\n");
    exit(1);
  }

  return p;
}

bool
InitPython()
{
  // Init the engine
  Py_Initialize();

  // Setup the Silk API
  Py_InitModule("silk", SilkPythonAPI);

    // Import stuff
  PyRun_SimpleString("import sys");
  PyRun_SimpleString("from silk import *");
  PyRun_SimpleString("sys.path.reverse()");  
  PyRun_SimpleString("sys.path.append(\".\")");
  PyRun_SimpleString("sys.path.reverse()");  

  // Done!
  return true;
}

PyObject*
ImportScript(const char *ScriptName)
{
  // Setup the name 
  PyObject *PyScriptName = PyString_FromString(ScriptName);

  // Import the script
  PyObject *PyScript = PyImport_Import(PyScriptName);
  Py_DECREF(PyScriptName);

  if(!PyScript)
  {
    fprintf(stderr, "Error: Failed to load python script \"%s\"\n", ScriptName);
    PyErr_Print();
    return NULL;
  }

  // Return script
  return PyScript;
}

// Returns ERROR_PTR on error
PyObject*
InvokeFunc(PyObject *PyScript, const char *FuncName, PyObject *PyArgs)
{
  // Find the method
  PyObject *PyFunc = PyObject_GetAttrString(PyScript, FuncName); // New ref
  PyObject *PyRet = NULL;

  // If the method does not exist, ignore it
  if(PyFunc && PyCallable_Check(PyFunc))
  {
    // It exists and it's callable
    // Just call it since it takes no arguments
    PyRet = PyObject_CallObject(PyFunc, PyArgs);
    if(!PyRet)
    {
      // Clean
      Py_DECREF(PyFunc);

      // Error
      fprintf(stderr, "Error: Function call to \"%s\" failed!\n", FuncName);
      PyErr_Print();

      // Return
      return (PyObject*)ERROR_PTR;
    }
  }
  Py_XDECREF(PyFunc);

  // Return
  return PyRet;
}

void 
InteractiveMode()
{
  // Read a command, execute it
  // TODO: Make this more usable, it can be used to create tunnels currently
  // and nothing more

  printf("Py> ");

  char Line[256];
  if(fgets(Line, sizeof(Line), stdin))
  {
    // Execute the line
    PyRun_SimpleString(Line);
  }
}

void ProxyMode()
{
  // For each UDP socket
  list<SockPair*>::iterator iudp = Globals::UDPConnections.begin();
  for(; iudp != Globals::UDPConnections.end();)
  {
    // SockPair
    SockPair *Pair = *iudp;

    // Tunnel ?
    if(!HandleTunnel(Pair, SILK_UDP, SILK_CLIENT))
    {
      // Remove
      iudp = Globals::UDPConnections.erase(iudp);
      continue;
    }

    if(!HandleTunnel(Pair, SILK_UDP, SILK_SERVER))
    {
      // Remove
      iudp = Globals::UDPConnections.erase(iudp);
      continue;
    }
    
    iudp++;
  }

  // For each server, check if any new connection is not spawned
  list<SockServ*>::iterator itcp = Globals::TCPServers.begin();
  for(; itcp != Globals::TCPServers.end(); itcp++)
  {
    SockServ *Serv = *itcp;

    // Anything to accept ?
    NetSock *NewClient = Serv->Server->Accept();
    if(NewClient)
      HandleAcceptTCP(NewClient, Serv);
  }

  // For each connection
  list<SockPair*>::iterator itcpc = Globals::TCPConnections.begin();
  while(itcpc != Globals::TCPConnections.end())
  {
    // SockPair
    SockPair *Pair = *itcpc;

    // Client
    if(!HandleTunnel(Pair, SILK_TCP, SILK_CLIENT))
    {
      // Remove
      itcpc = Globals::TCPConnections.erase(itcpc);
      continue;
    }

    // Server
    if(!HandleTunnel(Pair, SILK_TCP, SILK_SERVER))
    {
      // Remove
      itcpc = Globals::TCPConnections.erase(itcpc);
      continue;
    }

    // Iter!
    itcpc++;
  }
}

int CreateTCPProxy(
  int Type,
  char *ServBindHost, int ServBindPort,
  char *DstHost, int DstPort)
{
  // Create listen sock
  SockServ *Serv = new SockServ;
  memset(Serv, 0, sizeof(SockServ));

  Serv->Server = new NetSock;
  if(!Serv->Server->Listen((unsigned short)ServBindPort, ServBindHost))
  {
    fprintf(stderr, "Error: Could not create listen socket\n");
    delete Serv->Server;
    delete Serv;
    return -1;
  }

  // Change this to async
  Serv->Server->SetMode(NetSock::ASYNCHRONIC);

  // Get new handle
  int Handle = Globals::HandleCounter++;

  // Copy the rest of the data
  Serv->DstHost = xstrdup(DstHost);
  Serv->DstPort = DstPort;
  Serv->Type = Type;

  // Add to the list
  Globals::TCPServers.push_back(Serv);

  // Info
  printf("[id=%i] Tunnel created (listen=%s:%u, dst=%s:%u)\n",
      Handle,
      ServBindHost, ServBindPort,
      DstHost, DstPort);

  // Return
  return Handle;
}


int
CreateUDPProxy(int Type,
  char* SrcBindHost, int SrcBindPort,
  char* DefaultSrcHost, int DefaultSrcPort,
  char* DstHost, int DstPort, 
  char* DstBindHost, int DstBindPort)
{
  // Create two UDP socks
  SockPair *Pair = new SockPair;
  memset(Pair, 0, sizeof(SockPair));

  // Server (OUTGOING SOCKET)
  Pair->Server = new NetSock;
  if(!Pair->Server->ListenUDP((unsigned short)DstBindPort, DstBindHost))
  {
    fprintf(stderr, "Error: Could not create UDP server socket\n");
    delete Pair->Server;
    delete Pair;
    return -1;
  }

  // Change this to async
  Pair->Server->SetMode(NetSock::ASYNCHRONIC);

  // Client (INCOMMING SOCKET)
  Pair->Client = new NetSock;
  if(!Pair->Client->ListenUDP((unsigned short)SrcBindPort, SrcBindHost))
  {
    fprintf(stderr, "Error: Could not create UDP client socket\n");
    delete Pair->Server;
    delete Pair->Client;    
    delete Pair;
    return -1;
  }

  Pair->Client->SetMode(NetSock::ASYNCHRONIC);

  // Get new handle
  int Handle = Globals::HandleCounter++;

  // Copy the rest of the data
  Pair->Handle = -1; // Unused in UDP
  Pair->ServHandle = Handle;
  Pair->Type = Type;
  Pair->UDPDstPort = DstPort;
  Pair->UDPLastSrcPort = DefaultSrcPort;
  snprintf(Pair->UDPDstHost, 255, "%s", DstHost);
  snprintf(Pair->UDPLastSrcHost, 255, "%s", DefaultSrcHost);  

  // Add to the list
  Globals::UDPConnections.push_back(Pair);

  // Info
  printf("[id=%i] UDP Tunnel created (srcbind=%s:%u, src=%s:%u, dstbind=%s:%u, dst=%s:%u)\n",
      Handle,
      SrcBindHost, SrcBindPort,
      DefaultSrcHost, DefaultSrcPort,
      DstBindHost, DstBindPort,
      DstHost, DstPort);

  // TODO Logging

  // Return
  return Handle;
}

bool
HandleAcceptTCP(NetSock *NewClient, SockServ *Serv)
{
  // Print some info
  printf("[id=%i] TCP Client connected (src=%s:%u)\n",
    Serv->Handle,
    NewClient->GetStrIP(), NewClient->GetPort());

  // Try to connect
  NetSock *NewDescConn = new NetSock;
  if(NewDescConn->Connect(Serv->DstHost, (unsigned short)Serv->DstPort))
  {
    // It works!

    // Set the mode
    NewDescConn->SetMode(NetSock::ASYNCHRONIC);
    NewClient->SetMode(NetSock::ASYNCHRONIC);
        
    // Create a new sock pair
    SockPair *Pair = new SockPair;
    memset(Pair, 0, sizeof(SockPair));

    Pair->Client = NewClient;
    Pair->Server = NewDescConn;
    Pair->Handle = Globals::ConnHandleCounter++;
    Pair->ServHandle = Serv->Handle;    
    Pair->Type = Serv->Type;

    // Add logging
    if((Pair->Type & SILK_LOG))
    {
      // Craft some name
      char FName[256];
      snprintf(FName, 256, "TCP_%u_%s_%i_to_%s_%i.silk",
          (unsigned int)time(NULL),
          Pair->Client->GetStrIP(), Pair->Client->GetPort(),
          Pair->Server->GetStrIP(), Pair->Server->GetPort());

      Pair->Log = fopen(FName, "w");
      if(!Pair->Log)
      {
        fprintf(stderr, "Error: Could not create log file \"%s\"\n", FName);
      }
      else
      {
        printf("[id=%i,%i] Log file \"%s\"\n", 
            Serv->Handle, Pair->Handle, FName);
      }          
    }

    // Make the python call
    if(Globals::PyScript)
    {
      PyObject *PyParams = Py_BuildValue("((ii)(si)(si))",
          Pair->ServHandle, Pair->Handle,
          Pair->Client->GetStrIP(), Pair->Client->GetPort(),
          Pair->Server->GetStrIP(), Pair->Server->GetPort());

      PyObject *PyRet = InvokeFunc(
          Globals::PyScript, "Silk_OnTCPConnected", PyParams);

      Py_XDECREF(PyParams);
      Py_XDECREF(PyRet);

      // TODO let the function decide on some things     

    }

    // Add to the list
    Globals::TCPConnections.push_back(Pair);

    printf("[id=%i,%i] Tunneling TCP %s:%u->%s:%u\n",
      Serv->Handle, Pair->Handle, 
      NewClient->GetStrIP(), NewClient->GetPort(),
       Serv->DstHost, Serv->DstPort);

  }
  else
  {
    // Failed to connect
    printf("[id=%i] Could not connect to TCP destination %s:%u\n",
      Serv->Handle,
      Serv->DstHost, Serv->DstPort);

    // Delete stuff
    NewClient->Disconnect();
    delete NewClient;
    delete NewDescConn;

    return false;
  }

  return true;
}

bool 
HandleTunnel(SockPair *Pair, int Family, bool Client)
{
  // Some buffer
  static unsigned char Buffer[1024*1024];
  int Ret;
  char UDPSrcIP[32]; // 16 is enough, 32 for safety
  int  UDPSrcPort = 0;

  // Check C->S data
  NetSock *SrcSock = Client ? Pair->Client : Pair->Server;
  NetSock *DstSock = Client ? Pair->Server : Pair->Client;

  if(Family == SILK_TCP)
  {
    Ret = SrcSock->Read(Buffer, sizeof(Buffer));
  }
  else
  {
    Ret = SrcSock->ReadUDP(Buffer, sizeof(Buffer), UDPSrcIP, (unsigned short*)&UDPSrcPort);

    if(Ret > 0)
    {
      /*printf("UDP Data from %s: %s:%i\n", Client ? "client" : "server", UDPSrcIP, UDPSrcPort);*/

      // Setup last IP/port if it's from client
      if(Client)
      {
        // Copy stuff
        snprintf(Pair->UDPLastSrcHost, 255, "%s", UDPSrcIP);
        Pair->UDPLastSrcPort = UDPSrcPort;
      }
    }
  }

  // Disconnected ? (Requires a connection in the first place, so it's TCP only)
  if(Ret == 0 && Family == SILK_TCP)
  {
    // Disconnected
    printf("[id=%i,%i] TCP Tunneled connection %s:%u->%s:%u disconnected (by %s)\n",
      Pair->ServHandle, Pair->Handle, 
      Pair->Client->GetStrIP(), Pair->Client->GetPort(),
      Pair->Server->GetStrIP(), Pair->Server->GetPort(),
      Client ? "client" : "server");

    // TODO Call python
    if(Globals::PyScript)
    {
      PyObject *PyParams = Py_BuildValue("((ii)(si)(si)i)",
          Pair->ServHandle, Pair->Handle,
          Pair->Client->GetStrIP(), Pair->Client->GetPort(),
          Pair->Server->GetStrIP(), Pair->Server->GetPort(),
          Client ? 0 : 1);
 
      PyObject *PyRet = InvokeFunc(
          Globals::PyScript, "Silk_OnTCPDisconnected", PyParams);

      Py_XDECREF(PyParams);
      Py_XDECREF(PyRet);
    }

    // Disconnect all
    Pair->Client->Disconnect();
    Pair->Server->Disconnect();

    if(Pair->Log)
      fclose(Pair->Log);

    delete Pair->Client;
    delete Pair->Server;

    delete Pair;

    // Remove
    return false;
  }
  else if(Ret == -1)
  {
    // No data
    return true;
  }

  unsigned char *DataToSend = Buffer;
  int DataCount = Ret;

  // Call python Silk_OnTCPData
  if(Globals::PyScript && (Pair->Type & SILK_INT))
  {
    // UDP / TCP ?
    PyObject *PyParams = NULL;
    PyObject *PyRet = NULL;

    // Setup globals used by PyAPI
    Globals::CurrentPacketData = Buffer;
    Globals::CurrentPacketLen  = Ret;

    if(Family == SILK_TCP)
    {
      PyParams = Py_BuildValue("((ii)(si)(si)is#)",
        Pair->ServHandle, Pair->Handle,
        Pair->Client->GetStrIP(), Pair->Client->GetPort(),
        Pair->Server->GetStrIP(), Pair->Server->GetPort(),
        Client ? 0 : 1,
        Buffer, Ret);

      PyRet = InvokeFunc(
         Globals::PyScript, "Silk_OnTCPData", PyParams);
    }
    else if(Family == SILK_UDP)
    {
      PyParams = Py_BuildValue("(i(si)(sisi)is#)",
        Pair->ServHandle,
        Pair->UDPLastSrcHost, Pair->UDPLastSrcPort,
        Pair->UDPDstHost, Pair->UDPDstPort,
        Pair->Server->GetStrBindIP(), Pair->Server->GetBindPort(),
        Client ? 0 : 1,
        Buffer, Ret);

      PyRet = InvokeFunc(
         Globals::PyScript, "Silk_OnUDPData", PyParams);
    }

    if(PyRet && PyRet != Py_None && PyString_Check(PyRet))
    {
      // Data!
      Py_ssize_t Length = 0;
      PyString_AsStringAndSize(PyRet, (char**)&DataToSend, &Length);
      DataCount = (int)Length;
    }

    Py_XDECREF(PyParams);
    Py_XDECREF(PyRet);
  }

  // Log
  if((Pair->Type & SILK_LOG) && Pair->Log)
  {
    // TODO: for UDP, add address login
    PacketLog PLog;
    memset(&PLog, 0, sizeof(PLog));

    PLog.Type       = Client ? 1 : 0;
    PLog.OrgLen     = Ret;

    if(DataToSend != Buffer)
    {
      // One of two things:
      // 1. Either the data size differs
      // 2. Or the data differs
      if(DataCount != Ret ||
         memcmp(DataToSend, Buffer, DataCount) != 0)
      {
        // Include it
        PLog.OutPacketIncluded = true;
        PLog.OutLen = DataCount;
      }
    }

    // Write PacketLog
    fwrite(&PLog, 1, sizeof(PLog), Pair->Log);

    // Write Org Packet
    fwrite(Buffer, 1, Ret, Pair->Log);

    // Write Out Packet
    if(PLog.OutPacketIncluded)
      fwrite(DataToSend, 1, DataCount, Pair->Log);

    fflush(Pair->Log);
  }

  // TODO Verbose info

  // Some data came here, send it back
  DstSock->SetMode(NetSock::SYNCHRONIC);

  // TCP/UDP ?
  if(Family == SILK_TCP)
  {
    DstSock->Write(DataToSend, DataCount);
  }
  else if(Family == SILK_UDP)
  {
    if(Client)
    {
      // Send it to server
      DstSock->WriteUDP(Pair->UDPDstHost, Pair->UDPDstPort, DataToSend, DataCount);
    }
    else
    {
      // Send it to last client
      DstSock->WriteUDP(Pair->UDPLastSrcHost, Pair->UDPLastSrcPort, DataToSend, DataCount);      }
  }

  DstSock->SetMode(NetSock::ASYNCHRONIC);    


  return true;  
}

#ifdef __unix__
void 
HandleCtrlC(int SigNo)
#else
BOOL WINAPI
HandleCtrlC(DWORD CtrlType)
#endif
{
#ifdef __unix__
  UNUSED SigNo;
#else
  puts("in");
  if(CtrlType != CTRL_C_EVENT)
    return FALSE;
  puts("over");
#endif
  puts("x");

  // Interactive mode?
  if(Globals::InteractiveMode)
  {
    Globals::TheEnd = true;
  }
  else
  {
    Globals::InteractiveMode = true;
  }

#ifdef _WIN32
  return TRUE;
#endif
}

void
PrintHex(void *pdata, int len, int from, int to)
{
  unsigned char *data = (unsigned char*)pdata;
  int i;

  for(i = 0; i < len;)
  {
    char Line[256], *p, *r;
    memset(Line, ' ', 256);
    Line[79] = '\0';

    p = Line + 1;
    r = Line + 52 + 1;


    int j;
    for(j = 0; j < 16 && i < len; i++, p += 3, j++, r++)
    {
      if(from == i) p[-1] = '>';
      if(to == i)   p[2]  = '<';


      p[0] = "0123456789abcdef"[ (data[i] >> 4) & 0xf ];
      p[1] = "0123456789abcdef"[ data[i] & 0xf ];

      r[0] = '.';
      if(data[i] >= ' ' && data[i] <= '~')
        r[0] = data[i];

      if((i + 1) % 16 == 8) p++;
      if((i + 1) % 16 == 4) p++;
      if((i + 1) % 16 == 12) p++;
    }
    puts(Line);
  }
}


// ------------------------------------------------------------------
// Silk API
// ------------------------------------------------------------------
PyObject*
SilkAPI_Version(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Just return the version
  return Py_BuildValue("s", SILK_VERSION);
}

PyObject* 
SilkAPI_UDP(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Just return 1
  return Py_BuildValue("i", SILK_UDP);
}

PyObject*
SilkAPI_TCP(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Just return 0
  return Py_BuildValue("i", SILK_TCP);
}

PyObject*
SilkAPI_PASS(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Just return 0
  return Py_BuildValue("i", SILK_PASS);
}

PyObject*
SilkAPI_INT(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Just return 0
  return Py_BuildValue("i", SILK_INT);
}

PyObject*
SilkAPI_LOG(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Just return 0
  return Py_BuildValue("i", SILK_LOG);
}

PyObject* 
SilkAPI_Go(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Turn off InteractiveMode
  Globals::InteractiveMode = false;

  // Return None
  return Py_BuildValue("");
}

PyObject*
SilkAPI_Stop(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;  

  // Turn on InteractiveMode
  Globals::InteractiveMode = true;

  // Return None
  return Py_BuildValue("");
}

PyObject*
SilkAPI_Create(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;  

  // Args
  int Family = -1; // 0, 1
  char *ServBindHost = NULL;
  int   ServBindPort = -1;
  char *DstHost = NULL;
  int   DstPort = -1;
  char *DstBindHost = NULL;
  int   DstBindPort = -1;
  char *DefaultSrcHost = NULL;
  int   DefaultSrcPort = -1;
  int Type = 0;

  // Parse
  if(!PyArg_ParseTuple(args, "i(sizi)(sizi)|i",
        &Family,
        &ServBindHost,
        &ServBindPort,
        &DefaultSrcHost,
        &DefaultSrcPort,
        &DstHost,
        &DstPort,
        &DstBindHost,
        &DstBindPort,
        &Type))
    return NULL;

  // OK, start to play
  int Handle = -1;
  switch(Family)
  {
    case SILK_TCP:
      Handle = CreateTCPProxy(Type,
          ServBindHost, ServBindPort,
          DstHost, DstPort);
      break;

    case SILK_UDP:
      Handle = CreateUDPProxy(Type,
          ServBindHost, ServBindPort,
          DefaultSrcHost, DefaultSrcPort,
          DstHost, DstPort, 
          DstBindHost, DstBindPort);
      break;

    default:
      fprintf(stderr, "Error: Unknown family, could not create proxy\n");
      return NULL;
  }

  if(Handle == -1)
    return NULL;

  // Return None
  return Py_BuildValue("i", Handle);
}

PyObject*
SilkAPI_Destroy(PyObject *self, PyObject *args)
{
  UNUSED self;

  // Parse
  int Handle;
  if(!PyArg_ParseTuple(args, "i", &Handle))
    return NULL;

  // Find handle
  bool Found = false;

  list<SockServ*>::iterator i = Globals::TCPServers.begin();
  for(; i != Globals::TCPServers.end();)
  {
    SockServ *Serv = *i;

    // Remove if it's this
    if(Serv->Handle == Handle)
    {
      Found = true;
      delete Serv->Server;
      Globals::TCPServers.erase(i);
      break;
    }

    i++;
  }

  if(!Found)
  {
    fprintf(stderr,"Error: Handle \"%i\" not found\n", Handle);
    return NULL;
  }

  // Return None
  return Py_BuildValue("");
}

PyObject*
SilkAPI_Exit(PyObject *self, PyObject *args)
{
  UNUSED self;
  UNUSED args;

  // Done!
  Globals::TheEnd = true;

  return Py_BuildValue("");
}

PyObject*
SilkAPI_PrintHex(PyObject *self, PyObject *args)
{
  UNUSED self;

  // Parse
  char *Data;
  int DataLen;

  // Parse args
  if(!PyArg_ParseTuple(args, "z#", &Data, &DataLen))
    return NULL;

  // Cool, now test it
  if(Data == NULL)
  {
    PrintHex(Globals::CurrentPacketData, Globals::CurrentPacketLen, -1, -1);
  }
  else
  {
    PrintHex(Data, DataLen, -1, -1);
  }

  // Return None
  return Py_BuildValue("");
}

PyObject* 
SilkAPI_Fuzz(PyObject *self, PyObject *args)
{
  UNUSED self;

  char *Data = NULL;
  int DataLen = -1;
  int StartOffset = -1;
  int FuzzLen = -1;
  char *DataCopy;

  // Parse args
  if(!PyArg_ParseTuple(args, "s#ii", &Data, &DataLen, &StartOffset, &FuzzLen))
    return NULL;
  
  PyObject *PyStr = PyString_FromStringAndSize(Data, DataLen); // This should make a copy
  DataCopy = PyString_AsString(PyStr);

  // Chng Off leter to 'a'
  int Place = StartOffset + rand() % (DataLen - StartOffset - FuzzLen);
  int i;

  for(i = 0; i < FuzzLen; i++)
    DataCopy[Place + i] = rand() % 256;

  return PyStr;
}

PyObject* 
SilkAPI_Rand(PyObject *self, PyObject *args)
{
  UNUSED self;

  int Range = -1;

  // Parse args
  if(!PyArg_ParseTuple(args, "i", &Range))
    return NULL;
  
  return Py_BuildValue("i", rand() % Range);
}

PyObject*
SilkAPI_ReplayLoad(PyObject *self, PyObject *args)
{
  UNUSED self;

  char *FileName = NULL;
  int CachePackets = 0;

  // Parse args
  if(!PyArg_ParseTuple(args, "s|i", &FileName, &CachePackets))
    return NULL;

  // Try to open file
  


  return NULL;
}

PyObject*
SilkAPI_ReplayClose(PyObject *self, PyObject *args)
{
  UNUSED self;

  int Handle = -1;

  // Parse args
  if(!PyArg_ParseTuple(args, "i", &Handle))
    return NULL;

  // Find the handle
  // TODO Move this to the 


  return NULL;
}

PyObject*
SilkAPI_ReplayGetPacket(PyObject *self, PyObject *args)
{
  UNUSED self;

  return NULL;
}

